package gridcomp.manager;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import gridcomp.AutonomicClientManager;
import gridcomp.AutonomicServerManager;
import gridcomp.Constants;
import gridcomp.controller.AutonomicController;
import gridcomp.controller.BeskeAttributeController;
import gridcomp.log.Loggers;
import gridcomp.manager.beans.QoSEvent;
import gridcomp.manager.beans.QoSItemBean;
import gridcomp.manager.operations.ManagerCommand;
import gridcomp.manager.operations.ManagerOperation;
import gridcomp.operation.AbstractAutonomicOperation;

import org.apache.log4j.Logger;
import org.drools.FactHandle;
import org.drools.RuleBase;
import org.drools.RuleBaseFactory;
import org.drools.StatefulSession;
import org.drools.compiler.DroolsParserException;
import org.drools.compiler.PackageBuilder;
import org.drools.rule.Package;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.proactive.RunActive;
import org.objectweb.proactive.Body;
import org.objectweb.proactive.Service;
import org.objectweb.proactive.core.util.wrapper.GenericTypeWrapper;

import static gridcomp.Constants.MANAGER_CLIENT_ITF_TO_AUTONOMIC_CONTROLLER;
import static gridcomp.Constants.MANAGER_SLEEP_TIME;
public abstract class AbstractAutonomicManager implements RunActive, 
														  AutonomicServerManager, 
														  AutonomicClientManager,
														  BindingController {

	private boolean isManagerInitialized = false;
	protected static final Logger logger = Logger.getLogger(Loggers.MANAGER);
	protected long sleepTime = MANAGER_SLEEP_TIME;
	protected AutonomicController managedAutonomicController = null;
	protected AutonomicClientManager violationHandler; 
 	protected BeskeAttributeController bac = null;
 	protected static final String NAME_PREFIX = "slave-";
	protected StatefulSession wm = null;
	protected Component owner;
    // the rule engine has its own rulebase
	private RuleBase  ruleBase = RuleBaseFactory.newRuleBase();

//	private Map managerGlobals = new HashMap();
	// every manager should have its own of allowedOperation.
    private HashMap<ManagerOperation, ManagerCommand> allowedOperation = new HashMap<ManagerOperation, ManagerCommand>();	
    /* every manager should be able to handle the method monitoring  in its own way
     * So these addMethodToMonitor and removeMethodToMonitor are declared abstract.
     */
    public abstract void addMethodToMonitor(String data);
	public abstract void removeMethodToMonitor(String data);
	public abstract void sendEvent(GenericTypeWrapper<QoSEvent> ev);
	protected abstract void setupQoS();
		
	protected void initAutonomicManager(){
		isManagerInitialized = true;
		logger.info("AbstractAutonomicManager:: Manager initialized");
	}
	protected void addAutomomicOperation(ManagerOperation m, ManagerCommand op)
	{
		allowedOperation.put(m, op);
	}
	protected void removeAutomomicOperation(ManagerOperation m)
	{
		allowedOperation.remove(m);
	}
	protected ManagerCommand lookupOperation(ManagerOperation m)
	{
		return allowedOperation.get(m);
	}
	/* @param QosItemBean item
	 * @param update 
	 * you can add every item regarding Qos. QoSItem can be a single fact: 
	 * serviceTime, numberOfWorker...ecc.
	 * You can update a fact in the rule engine or simply add a new one
	 */
	@SuppressWarnings("unchecked")
	protected  void addQoSItem(QoSItemBean item, boolean update)
	{
		FactHandle fact = wm.getFactHandle(item);
		if ((!update) || (fact==null))
		{
			logger.info("Manager is adding the following QoS bean "+ item.getClass().getName());
			
			wm.insert(item);
		} else {
			logger.info("Manager is updating the following QoS bean "+ item.getClass().getName());
			wm.update(fact, item);
		}

	}
	@SuppressWarnings("unchecked")
	protected  void removeQoSItem(QoSItemBean item)
	{
		FactHandle fact = wm.getFactHandle(item);
		wm.retract(fact);
	}
   protected void enforcePerformanceContract ( )
	{
	   if (wm!=null)
		{
		logger.info("Detecting information from the enviroment");
		setupQoS();
		logger.info("Valuating information by fireAllRules()");
		wm.fireAllRules();
		}
	}
	
	// Method invocated to specify the performance contract
	public GenericTypeWrapper<Object> commitContract ( String qosContract )
	{	
		
		RuleBase ruleBase = readRule(qosContract);
		wm = ruleBase.newStatefulSession();
		return null;	
	}
	private Package createNewPackage(String rule)
	{
		PackageBuilder packBuilder = new PackageBuilder();
		try {
			packBuilder.addPackageFromDrl(new StringReader(rule));
		} catch (DroolsParserException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
		logger.info("Creating a rule package from the contract");
		Package pkg = packBuilder.getPackage();
		return pkg;
	}
	private RuleBase readRule(String rule)
	{
	
		try {
			Package pkg = createNewPackage(rule);
			if (pkg!=null)
				ruleBase.addPackage(pkg);
			else
			  logger.info("Error reading a rule");
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		return ruleBase;
	}
	// Method invocated by the lower level manager(s) raising violations
//	public abstract GenericTypeWrapper<Object> raiseViolation ( GenericTypeWrapper<Object> violationId );
	
	// Autonomic Manager Main Loop
	public void runActivity(Body body) {
		
	    logger.warn("Manager cycle starting ...");
		Service serv = new Service(body);
		logger.info("Soddisfo richieste pendenti");
		while(serv.hasRequestToServe()){
				serv.serveOldest();
		}
		
		//logger.warn("Inizializzo Manager");
		// Initializing Autonomic Manager
		if(!isManagerInitialized) 
			initAutonomicManager();
	        // logger.warn("Dormo k secondi");
		// Sleep "sleepTime" seconds 
		try {
			Thread.sleep(sleepTime);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		logger.info("Controlla contratto");
		// Enforcing Performance Contract
		try {
			
			enforcePerformanceContract ( );
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// bindFc method for the binding between manager and controller
 	public void bindFc(String itfName, Object itf) throws NoSuchInterfaceException, IllegalBindingException, IllegalLifeCycleException {
		
		if(itfName == null) throw new NoSuchInterfaceException("Interface name cannot be 'null'");
		if(itf == null) throw new IllegalBindingException("Interface binded cannot be 'null'");
		if (itfName.equals(gridcomp.Constants.MANAGER_CLIENT_ITF_TO_BESKELETON_CONTROLLER))
		{
			bac = (BeskeAttributeController) itf;
			logger.info("Binding beske-interface to class "+ itf.getClass().getName());
		}
		if(itfName.equals(MANAGER_CLIENT_ITF_TO_AUTONOMIC_CONTROLLER)){
			managedAutonomicController = (AutonomicController) itf;
			logger.info("AbstractAutonomicManager:: binding with the autonomic controller to class "+itf.getClass().getName());
		}else if(itfName.equals(RAISE_VIOLATION_ITF_NAME)){
			violationHandler = (AutonomicClientManager) itf;
			logger.info("AbstractAutonomicManager:: binding with the violation handler correctly performed");
		}
	}
	
 	// listFc method for the binding between manager and controller
	public String[] listFc() {
		return new String[]{MANAGER_CLIENT_ITF_TO_AUTONOMIC_CONTROLLER, gridcomp.Constants.MANAGER_CLIENT_ITF_TO_BESKELETON_CONTROLLER,
							RAISE_VIOLATION_ITF_NAME};
	}

	// lookupFc method for the binding between manager and controller
	public Object lookupFc(String itfName) throws NoSuchInterfaceException {
		
		if(itfName == null) throw new NoSuchInterfaceException("Interface name cannot be 'null'");
		if (itfName.equals(gridcomp.Constants.MANAGER_CLIENT_ITF_TO_BESKELETON_CONTROLLER))
		{
			Object result = bac;
			if (result!=null) return result;
		}
	
		if(itfName.equals(MANAGER_CLIENT_ITF_TO_AUTONOMIC_CONTROLLER)){
			Object result = managedAutonomicController;
			if(result != null) return result;
		}else if(itfName.equals(RAISE_VIOLATION_ITF_NAME)){
			Object result = violationHandler;
			if(result != null) return result;
		}
		
		throw new NoSuchInterfaceException(itfName);
	}

	// unbindFc method for the binding between manager and controller
	public void unbindFc(String itfName) throws NoSuchInterfaceException, IllegalBindingException, IllegalLifeCycleException {
		
		if(itfName == null) throw new NoSuchInterfaceException("Interface name cannot be 'null'");
		
		if(itfName.equals(MANAGER_CLIENT_ITF_TO_AUTONOMIC_CONTROLLER)){
			managedAutonomicController = null;
			logger.info("AbstractAutonomicManager:: unbinding with the autonomic controller correctly performed");
		}else if(itfName.equals(RAISE_VIOLATION_ITF_NAME)){
			violationHandler = null;
			logger.info("AbstractAutonomicManager:: unbinding with the violation handler correctly performed");
		}else throw new NoSuchInterfaceException(itfName);
	}
	
}
